@@ -3,9 +3,9 @@ use strict;
 use warnings;
 use 5.010;
 
-our $VERSION = '0.17';
+our $VERSION = '0.18';
 
-our $PANDOC_VERSION;   # a string like '1.16'
+our $PANDOC_VERSION;    # a string like '1.16'
 $PANDOC_VERSION ||= $ENV{PANDOC_VERSION};
 
 our %ELEMENTS = (
@@ -68,7 +68,7 @@ foreach (
 
 use Carp;
 use JSON qw(decode_json);
-use Scalar::Util qw(reftype);
+use Scalar::Util qw(blessed reftype);
 use Pandoc::Walker qw(walk);
 
 use parent 'Exporter';
@@ -132,12 +132,16 @@ sub Pandoc::Document::DefinitionPair::definitions { $_[0]->[1] }
 
 sub attributes($) {
     my ($attrs) = @_;
+    my $classes = $attrs->{classes} // [];
+    if ( defined $attrs->{class} ) {
+        unshift @$classes, grep { $_ ne '' } split qr/\s+/, $attrs->{class};
+    }
     return [
-        defined $attrs->{id}      ? $attrs->{id}      : '',
-        defined $attrs->{classes} ? $attrs->{classes} : [],
+        $attrs->{id} // '',
+        $classes,
         [
-            map { $_ => $attrs->{$_} }
-              grep { $_ ne 'id' and $_ ne 'classes' }
+            map { [ $_ => $attrs->{$_} ] }
+              grep { $_ !~ qr/^(id|class(es)?)$/ }
               keys %$attrs
         ]
     ];
@@ -159,14 +163,14 @@ sub citation($) {
 }
 
 sub pandoc_json($) {
-    state $ast_to_element = sub { # compile once, use repeatedly, keep private
+    state $ast_to_element = sub {   # compile once, use repeatedly, keep private
         my $class = 'Pandoc::Document::' . $_[0]->{t};
         if ( 'MetaMap' eq $_[0]->{t} ) {
             for my $v ( values %{ $_[0]->{c} } ) {
-                $_[1]->($v, $_[1]);
+                $_[1]->( $v, $_[1] );
             }
         }
-        return $class->new_from_ast($_[0])
+        return $class->new_from_ast( $_[0] );
     };
 
     shift if $_[0] =~ /^Pandoc::/;
@@ -182,13 +186,14 @@ sub pandoc_json($) {
     if ( reftype $ast eq 'ARRAY' ) {
         my $meta = $ast->[0]->{unMeta};
         for my $v ( values %$meta ) {
-            $v = $ast_to_element->($v, $ast_to_element);
+            $v = $ast_to_element->( $v, $ast_to_element );
         }
         $ast = Document( $meta, $ast->[1] );
     }
     elsif ( reftype $ast eq 'HASH' and $ast->{t} ) {
+
         # $ast = element( $ast->{t}, $ast->{c} );
-        $ast = $ast_to_element->($ast, $ast_to_element)
+        $ast = $ast_to_element->( $ast, $ast_to_element );
     }
 
     walk $ast, $ast_to_element;
@@ -204,7 +209,6 @@ sub pandoc_json($) {
     use strict;
     our $VERSION = '0.04';
     our @ISA     = ('Pandoc::Document::Element');
-    sub TO_JSON     { [ @{ $_[0] } ] }
     sub name        { 'Document' }
     sub meta        { $_[0]->[0]->{unMeta} }
     sub content     { $_[0]->[1] }
@@ -218,15 +222,47 @@ sub pandoc_json($) {
     use warnings;
     our $VERSION = $Pandoc::Document::VERSION;
     use JSON ();
-    use Scalar::Util qw(reftype);
+    use Scalar::Util qw(reftype blessed);
     use Pandoc::Walker ();
+    use subs qw(walk query transform);    # Silence syntax warnings
 
     sub to_json {
         JSON->new->utf8->convert_blessed->encode( $_[0] );
     }
-    sub TO_JSON { return { %{ $_[0] } } }
-    sub name    { $_[0]->{t} }
-    sub content { $_[0]->{c} }
+
+    sub TO_JSON {
+
+        # Run everything thru this method so arrays/hashes are cloned
+        # and objects without TO_JSON methods are stringified.
+        # Required to ensure correct scalar types for Pandoc.
+
+# There is no easy way in Perl to tell if a scalar value is already a string or number,
+# so we stringify all scalar values and numify/boolify as needed afterwards.
+
+        my ( $ast, $maybe_blessed ) = @_;
+        if ( $maybe_blessed && blessed $ast ) {
+            return $ast if $ast->can('TO_JSON');    # JSON.pm will convert
+                 # may have overloaded stringification! Should we check?
+                 # require overload;
+              # return "$ast" if overload::Method($ast, q/""/) or overload::Method($ast, q/0+/);
+              # carp "Non-stringifiable object $ast";
+            return "$ast";
+        }
+        elsif ( 'ARRAY' eq reftype $ast ) {
+            return [ map { ref($_) ? TO_JSON( $_, 1 ) : "$_"; } @$ast ];
+        }
+        elsif ( 'HASH' eq reftype $ast ) {
+            my %ret = %$ast;
+            while ( my ( $k, $v ) = each %ret ) {
+                $ret{$k} = ref($v) ? TO_JSON( $v, 1 ) : "$v";
+            }
+            return \%ret;
+        }
+        else { return "$ast" }
+    }
+
+    sub name        { $_[0]->{t} }
+    sub content     { $_[0]->{c} }
     sub is_document { 0 }
     sub is_block    { 0 }
     sub is_inline   { 0 }
@@ -243,14 +279,14 @@ sub pandoc_json($) {
         if ( $_[0]->name =~ /^(Str|Code|Math|MetaString)$/ ) {
             return $_[0]->content;
         }
-        elsif ( $_[0]->name =~ /^(LineBreak|Space|MetaString)$/ ) {
+        elsif ( $_[0]->name =~ /^(LineBreak|SoftBreak|Space|MetaString)$/ ) {
             return ' ';
         }
         join '', @{
             $_[0]->query(
                 {
-                    'Str|Code|Math|MetaString'   => sub { $_->content },
-                    'LineBreak|Space' => sub { ' ' },
+                    'Str|Code|Math|MetaString'  => sub { $_->content },
+                    'LineBreak|Space|SoftBreak' => sub { ' ' },
                 }
             );
         };
@@ -345,35 +381,91 @@ sub pandoc_json($) {
     package Pandoc::Document::LinkageRole;
     our $VERSION = $PANDOC::Document::VERSION;
 
-    for my $Element ( qw[ Link Image ] ) {
-        no strict 'refs'; #no critic
+    for my $Element (qw[ Link Image ]) {
+        no strict 'refs';    #no critic
         unshift @{"Pandoc::Document::${Element}::ISA"}, __PACKAGE__; # no critic
     }
 
-    sub url                   { $_[0]->{c}->[-1][0] }
-    sub title                 { $_[0]->{c}->[-1][1] }
-    
+    sub url   { $_[0]->{c}->[-1][0] }
+    sub title { $_[0]->{c}->[-1][1] }
+
     sub new_from_ast {
-        my ($class, $ast) = @_;
+        my ( $class, $ast ) = @_;
         if ( 2 == @{ $ast->{c} } ) {
+
             # prepend attributes to old-style ast
-            unshift @{ $ast->{c} }, ["",[],[]];
+            unshift @{ $ast->{c} }, [ "", [], [] ];
         }
         return bless $ast => $class;
     }
+}
 
-    sub TO_JSON {
-        my ( $self ) = @_;
-        my $ast = {%$self};
-        if (    $Pandoc::Elements::PANDOC_VERSION
-            and ($Pandoc::Elements::PANDOC_VERSION lt '1.16') )
-        {
-            # remove attributes from new-style ast
-            $ast->{c} = [ @{ $ast->{c} }[ 1, 2 ] ];
-        }
-        return $ast;
+# Special TO_JSON methods to coerce data to int/number/Boolean as appropriate
+# and to downgrade document model for Pandoc < 1.16
+
+sub Pandoc::Document::SoftBreak::TO_JSON {
+    if ( $Pandoc::Elements::PANDOC_VERSION
+        and ( $Pandoc::Elements::PANDOC_VERSION lt '1.16' ) ) {
+        return { t => 'Space', c => [] };
+    } else {
+        return { t => 'SoftBreak', c => [] };
     }
+}
 
+sub Pandoc::Document::LinkageRole::TO_JSON {
+    my $ast = Pandoc::Document::Element::TO_JSON( $_[0] );
+    if ( $Pandoc::Elements::PANDOC_VERSION
+        and ( $Pandoc::Elements::PANDOC_VERSION lt '1.16' ) )
+    {
+        # remove attributes from new-style ast
+        $ast->{c} = [ @{ $ast->{c} }[ 1, 2 ] ];
+    }
+    return $ast;
+}
+
+sub Pandoc::Document::Header::TO_JSON {
+    my $ast = Pandoc::Document::Element::TO_JSON( $_[0] );
+
+    # coerce heading level to int
+    $ast->{c}[0] = int( $ast->{c}[0] );
+    return $ast;
+}
+
+sub Pandoc::Document::OrderedList::TO_JSON {
+    my $ast = Pandoc::Document::Element::TO_JSON( $_[0] );
+
+    # coerce first item number to int
+    $ast->{c}[0][0] = int( $ast->{c}[0][0] );
+    return $ast;
+}
+
+sub Pandoc::Document::Table::TO_JSON {
+    my $ast = Pandoc::Document::Element::TO_JSON( $_[0] );
+
+    # coerce column widths to numbers (floats)
+    $_ += 0 for @{ $ast->{c}[2] };    # faster than map
+    return $ast;
+}
+
+sub Pandoc::Document::MetaBool::TO_JSON {
+    return {
+        t => 'MetaBool',
+
+        # coerce Bool value to JSON Boolean object
+        c => $_[0]->{c} ? JSON::true() : JSON::false(),
+    };
+}
+
+sub Pandoc::Document::Cite::TO_JSON {
+    my $ast = Pandoc::Document::Element::TO_JSON( $_[0] );
+    for my $citation ( @{ $ast->{c}[0] } ) {
+        for my $key (qw[ citationHash citationNoteNum ]) {
+
+            # coerce to int
+            $citation->{$key} = int( $citation->{$key} );
+        }
+    }
+    return $ast;
 }
 
 1;
@@ -389,8 +481,8 @@ Pandoc::Elements - create and process Pandoc documents
 
 # STATUS
 
-[![Build Status](https://travis-ci.org/nichtich/Pandoc-Elements.png)](https://travis-ci.org/nichtich/Pandoc-Elements)
-[![Coverage Status](https://coveralls.io/repos/nichtich/Pandoc-Elements/badge.png)](https://coveralls.io/r/nichtich/Pandoc-Elements)
+[![Build Status](https://travis-ci.org/nichtich/Pandoc-Elements.svg)](https://travis-ci.org/nichtich/Pandoc-Elements)
+[![Coverage Status](https://coveralls.io/repos/nichtich/Pandoc-Elements/badge.svg)](https://coveralls.io/r/nichtich/Pandoc-Elements)
 [![Kwalitee Score](http://cpants.cpanauthors.org/dist/Pandoc-Elements.png)](http://cpants.cpanauthors.org/dist/Pandoc-Elements)
 
 =end markdown
@@ -465,11 +557,13 @@ format.
 =head3 attributes { key => $value, ... }
 
 Maps a hash reference into an attributes list with id, classes, and ordered
-key-value pairs. The special keys C<id> and C<classes> are recognized but
-setting multi-value attributes or controlled order is not supported with this
-function. You can always manually create an attributes structure:
+key-value pairs. The special keys C<id> (string), C<classes> (array reference
+of class names), and C<class> (string with space-separated class names) are
+recognized but setting multi-value attributes or controlled order is not
+supported with this function. You can always manually create an attributes
+structure:
 
-    [ $id, [ @classes ], [ key => $value, ... ] ]
+    [ $id, [ @classes ], [ [ key => $value ], ... ] ]
 
 Elements with attributes (element accessor method C<attr>) also provide the
 accessor method C<id>, C<classes>, and C<class>. See L<Hash::MultiValue> for
@@ -481,7 +575,7 @@ A citation as part of document element L<Cite|/Cite> must be a hash reference
 with fields C<citationID> (string), C<citationPrefix> (list of L<inline
 elements|/INLINE ELEMENTS>) C<citationSuffix> (list of L<inline
 elements|/INLINE ELEMENTS>), C<citationMode> (one of C<NormalCitation>,
-C>AuthorInText>, C<SuppressAuthor>), C<citationNoteNum> (integer), and
+C<AuthorInText>, C<SuppressAuthor>), C<citationNoteNum> (integer), and
 C<citationHash> (integer). The helper method C<citation> can be used to
 construct such hash by filling in default values and using shorter field names
 (C<id>, C<prefix>, C<suffix>, C<mode>, C<note>, and C<hash>):
@@ -763,6 +857,12 @@ Soft line break
 
     SoftBreak
 
+Note that the C<SoftBreak> element was added in Pandoc 1.16 to as a matter of
+editing convenience to preserve line breaks (as opposed to paragraph breaks)
+from input source to output. If you are going to feed a document containing
+C<SoftBreak> elements to Pandoc E<lt> 1.16 you will have to set the package
+variable or environment variable C<PANDOC_VERSION> to 1.15 or below.
+
 =head3 Space
 
 Inter-word space
